import os
from pathlib import Path
from timeit import default_timer
from copy import deepcopy
import pickle
import numpy as np
import pandas as pd
from pprint import pp

from mcfnmr.config import default_pars, OUTDIR, DEBUG
from mcfnmr.core.mcf import mcf
from mcfnmr.demodata import binning
from mcfnmr.demodata import (
    mixNamesOLDB,
    compoundShortNames,
    scanNumbersMixes,
    scanNumberIndividualCompounds,
    intensityScalings,
    expectedConcentrationFactors,
)
from mcfnmr.demodata.loading import loadMetaboMinerSpectrum
from mcfnmr.utils.pointspectrum import makeRasterPointSpectrum

cache_dir = Path(OUTDIR) / "cache_demo_routines"
if not cache_dir.exists():
    os.mkdir(cache_dir)
    print(f"Created cache directory {cache_dir}")

figs_dir = OUTDIR / "figs"
figs_dir_oldb = figs_dir / "quantification"
figs_dir_MM = figs_dir / "classification MM"
if not figs_dir_oldb.exists():
    os.makedirs(figs_dir_oldb)
    print(f"Created directory {figs_dir_oldb}")
if not figs_dir_MM.exists():
    os.makedirs(figs_dir_MM)
    print(f"Created directory {figs_dir_MM}")


def makeTargetSpecification(pars, targetID, mixID=None):
    if mixID is None:
        mixID = targetID
    if not pars["pointTarget"]:
        nbin = pars.get("binning")
        if nbin is None:
            nbin = binning.get(mixID, None)
            if nbin is None:
                print("No nbin for", mixID)
        smoothstr = ", smooth%g" % pars["smoothing"] if pars["smoothing"] else ""
        binstr = ", (%dx%d)-bins" % (nbin)
        targetSpecification = targetID + "(raster%s%s)" % (smoothstr, binstr)
    else:
        targetSpecification = targetID + "(peaklist)"
    return targetSpecification


def makeMCFResultFilename(pars, mixID):
    if "specificTargetSpecID" not in pars:
        pars["specificTargetSpecID"] = makeTargetSpecification(
            pars, pars["targetSpecID"], mixID
        )
    targetID = pars["specificTargetSpecID"]
    compoundFit = "isolated" if pars["isolated_fit"] else "joint"
    noiseFilterLevelStr = (
        "_noiseFilter%g" % pars["noiseFilterLevel"] if pars["noiseFilterLevel"] else ""
    )
    ar = pars["assignmentRadius"]
    if "previousARSteps" in pars:
        prevArSteps = ["%g" % s for s in sorted(pars["previousARSteps"])]
        arStr = f"_%gon%s" % (float(ar), "[" + (",".join(prevArSteps)) + "]")
    else:
        arStr = "_ar%g" % ar
    libID = pars["libID"]
    fn = (
        f"mcfResult_{compoundFit}_in_{targetID}_lib_{libID}{noiseFilterLevelStr}{arStr}"
    )
    fn = fn + ".pickle"
    return fn


def collect_and_serialize_MM_grid_spectra(mixIDs, pars):
    mixSpecs = dict()
    for mixID in mixIDs:
        targetRasterSpec = loadMetaboMinerSpectrum(
            mixID, denoising_pars={"smoothRadius": pars["smoothing"]}
        )
        nbins = pars.get("binning", None)
        if nbins is None:
            nbins = binning[mixID]
        mixSpecs[mixID] = makeRasterPointSpectrum(
            targetRasterSpec,
            nbin=nbins,
            signalThreshold=pars["noiseFilterLevel"],
        )
    return mixSpecs


def runMetaboMinerFits(
    mixIDs, base_pars, lib, mixSpecs, returnAllResults=False, load=True, verb=1
):
    """
    Fit Metabominer data and compute precision/recall/F score
    """
    incrementalFit = base_pars["incrementalFit"]
    pointTarget = base_pars["pointTarget"]

    results = {}
    for mixID in mixIDs:
        pars = deepcopy(base_pars)  # protect from mutation
        pars["targetSpecID"] = mixID
        results[mixID] = {}
        targetPointSpec = mixSpecs[mixID]
        if pointTarget:
            # working with point spectra as targets
            # load regions from raster spectrum
            if mixID[-4:] == "(dp)":
                # Deeppicker variant of peaklist
                targetRasterSpec = loadMetaboMinerSpectrum(
                    mixID[:-4], denoising_pars={"smoothRadius": pars["smoothing"]}
                )
            else:
                targetRasterSpec = loadMetaboMinerSpectrum(
                    mixID, denoising_pars={"smoothRadius": pars["smoothing"]}
                )

            pars["binning"] = (1, 1)
            targetPointSpec.regions = targetRasterSpec.regions
        else:
            print(
                f"runMetaboMinerFits(): Binning for mix {mixID}: {targetPointSpec.binning}"
            )

        pars["specificTargetSpecID"] = makeTargetSpecification(
            pars, pars["targetSpecID"], mixID
        )

        # Generate or load MCF results
        if incrementalFit:
            # Initial setup for AR steps.
            # This triggers incremental fit in singleRun()
            pars["previousARSteps"] = []
            pars["reservedSinkCapacity"] = np.zeros_like(targetPointSpec.weights)
            prevResult, prevSavefn = None, None
            for ar in sorted(pars["arSpan"]):
                pars["assignmentRadius"] = ar
                savefn = cache_dir / makeMCFResultFilename(pars, mixID)
                result = singleRun(
                    pars, lib, targetPointSpec, load=load, verb=verb, savefn=savefn
                )
                inflowsY = getattr(result, "inflowsY", None)
                loadedResult = inflowsY is None
                if not loadedResult:
                    # Update reserved capacity, if not loading results
                    incrementalMCFResultsUpdate(result, inflowsY, prevResult)
                    # Delete inflowsY and residualsY from result to save storage.
                    # We will only need the reserved capacity at the next
                    # larger ar-value to extend an incremental series
                    result.inflowsY = None
                    result.residualsY = None
                    updateSavedMCFResult(result, savefn=savefn)
                    if prevResult:
                        # reservedCapacity not needed for non-maximal ar
                        prevResult.reservedSinkCapacity = None
                        updateSavedMCFResult(prevResult, savefn=prevSavefn)
                prevSavefn = savefn
                results[mixID][ar] = result
                pars["previousARSteps"].append(float(ar))
                pars["reservedSinkCapacity"] = getattr(
                    result, "reservedSinkCapacity", None
                )
                prevResult = result
        else:
            savefn = cache_dir / makeMCFResultFilename(pars, mixID)
            results[mixID] = singleRun(
                pars, lib, targetPointSpec, load=load, verb=verb, savefn=savefn
            )

    if not returnAllResults and incrementalFit:
        results = {
            mixID: results_i[max(results_i.keys())]
            for mixID, results_i in results.items()
        }
    return results


def updateSavedMCFResult(result, savefn):
    with open(savefn, "wb") as f:
        pickle.dump(result, f)
    print(f"Saved updated result to '{savefn}'")


def prepareParameters(specDomain, targetSpecID, libID, **kwargs):
    pars = deepcopy(default_pars)

    # Whether to use the peak list dat of the target (or the raster data)
    # (Only effective if both are available)
    assert "pointTarget" in pars

    pars["specDomain"] = specDomain
    pars["targetSpecID"] = targetSpecID
    pars["libID"] = libID

    pars.update(kwargs)
    return pars


def totalWeights(lib):
    weights = {}
    for c, spec in lib.items():
        w = sum(spec.weights)
        weights[c] = w
    return weights


def incrementalMCFResultsUpdate(res, inflowsY, prev):
    if prev is None:
        # res must be the first result in a sequence of
        # assignment radii if we are here
        res.reservedSinkCapacity = inflowsY
        res.totalReserved = np.sum(res.reservedSinkCapacity)
        return

    res.reservedSinkCapacity = prev.reservedSinkCapacity + inflowsY
    res.totalReserved = np.sum(res.reservedSinkCapacity)

    res.originalWeightY = prev.originalWeightY

    totalAbsorbed = res.totalAbsorption + prev.totalAbsorption
    absorptionCost = (
        res.totalAbsorption * res.absorptionCost
        + prev.totalAbsorption * prev.absorptionCost
    ) / totalAbsorbed
    res.totalAbsorbed = totalAbsorbed
    res.absorptionCost = absorptionCost

    assigned = res.assigned + prev.assigned
    assignedIx = assigned > 0
    assignmentCosts = (
        res.assigned * res.assignmentCosts + prev.assigned * prev.assignmentCosts
    )
    assignmentCosts[assignedIx] /= assigned[assignedIx]
    res.assigned = assigned
    res.assignmentCosts = assignmentCosts
    res.specificCosts[assignedIx] = assignmentCosts[assignedIx] / assigned[assignedIx]

    res.inflowsC += prev.inflowsC
    res.totalAbsorption += prev.totalAbsorption
    res.totalAssignedFlow += prev.totalAssignedFlow

    res.totalAbsorptionCost = res.totalAbsorption * res.absorptionCost
    res.totalAssignmentCost = sum(res.assignmentCosts)
    res.totalCost = res.totalAbsorptionCost + res.totalAssignmentCost


def singleRun(pars, lib, targetPointSpec, savefn, verb=1, load=True):
    pars = deepcopy(pars)
    targetPointSpec = deepcopy(targetPointSpec)
    targetPointSpec.weights = np.maximum(targetPointSpec.weights, 0.0)

    # This indicates, that this fit is a step in a assignment
    # radius increase / sink reservation series
    incrementalFit = "previousARSteps" in pars
    if incrementalFit and not (load and savefn.exists()):
        prevSteps = pars["previousARSteps"]
        if len(prevSteps) > 0:
            assert "reservedSinkCapacity" in pars
            # We have previous assignments → reserve the sinks
            targetPointSpec.weights -= pars["reservedSinkCapacity"]
            reserved_precision_tolerance = 1e-16 * max(targetPointSpec.weights)
            if min(targetPointSpec.weights) < 0.0:
                if min(targetPointSpec.weights < -reserved_precision_tolerance):
                    # Negativity exceeds tolerated round-off errors
                    raise Exception(
                        "Incremental fit encountered reserved capacity exceeding target capacity."
                    )
                else:
                    # Ignore negativity, since it is probably due to a rounding error
                    targetPointSpec.weights = np.maximum(targetPointSpec.weights, 0.0)

        if DEBUG > 0:
            rc = pars.get("reservedSinkCapacity", [0.0])
            print(
                "\nIncremental fit:\n   total reserved capacity: %g\n    min/max: %g/%g\n"
                % (sum(rc), min(rc), max(rc))
            )

    resolveYinResult = incrementalFit

    tic = default_timer()
    mcfResult = mcf(
        target_spectrum=targetPointSpec,
        library=lib,
        assignment_radius=pars["assignmentRadius"],
        absorption_cost=pars["absorptionCost"],
        target_id=pars["specificTargetSpecID"],
        lib_id=pars["libID"],
        target_regions=targetPointSpec.regions,
        isolated_fit=pars["isolated_fit"],
        resolveYinResult=resolveYinResult,
        load_dists=pars.get("load_dists", False),
        load=load,
        savefn=savefn,
    )
    print("Total time for MCF: %g secs\n" % (default_timer() - tic))
    return mcfResult


def filterDF(dfInit, filters):
    df = deepcopy(dfInit)
    for k, v in filters.items():
        df = df.loc[df[k] == v]
    return df


def makeResultsDF(data, originalWeights, intensityCorrection=True):
    # NOTE: Consider splitting this for purposes of MetaboMiner and OLDB-experiments
    # (handling both cases makes this difficult to understand)
    mcfResults = {}
    params = {}
    for d in data:
        res, pars = d["res"], d["pars"]
        fn = makeMCFResultFilename(pars, pars["targetSpecID"])
        if DEBUG > 0:
            if fn in mcfResults:
                raise Exception("makeResultsDF(): Duplicate run id '%s'" % fn)
        mcfResults[fn] = res
        params[fn] = pars

    runIDs = sorted(mcfResults.keys())

    dfCompounds = {
        "compound": [],
        "target": [],
        "lib": [],
        "assignment radius": [],
        "assigned": [],
        "absorbed": [],
        "specific cost": [],
        "concentration factor": [],
        "absorption factor": [],
        "conc.fctr. mismatch": [],
        "isolated fit": [],
        "incremental fit": [],
        "nPeaks": [],
        "intensity scaling": [],  # automatic intensity scaling
        "expected concentration factor": [],
    }
    dfTargets = {
        "target": [],
        "lib": [],
        "assignment radius": [],
        "assigned": [],
        "absorbed": [],
        "specific cost": [],
        "isolated fit": [],
        "originalWeightY": [],
    }

    for runID in runIDs:
        pars, results = params[runID], mcfResults[runID]
        targetID = pars["targetSpecID"]
        if targetID in mixNamesOLDB:
            targetName = mixNamesOLDB[targetID]
        elif targetID.split("_")[0] in compoundShortNames:
            compoundFullName = targetID.split("_")[0]
            targetName = compoundShortNames[compoundFullName]
        else:
            targetName = targetID
        if targetID in scanNumbersMixes:
            # Corrector for the mix concentration (all individual compounds are created by 8 scans,
            # mixes sometimes at differing numbers)
            nScansY = scanNumbersMixes[targetID]
        else:
            nScansY = scanNumberIndividualCompounds
        libID = pars["libID"]
        ar = pars["assignmentRadius"]
        isolated_fit = pars["isolated_fit"]
        incrFit = pars["incrementalFit"]

        if type(results) is not list:
            # Transform single result to data structure obtained if no absorption limits are specified
            results = [{"res": results, "absorptionLimit": None}]

        for resInfo in results:
            res = resInfo["res"]
            compounds = sorted(res.compounds.keys())
            if DEBUG > 0:
                missing = [c for c in compounds if not c in originalWeights]
                if missing:
                    print(
                        f"Compounds {missing}: no original weight found -> not in lib?"
                    )
                    print(libID)

            if DEBUG > 1:
                print("originalWeights (%d):" % len(originalWeights))
                pp(originalWeights)
                print("\n ################ \n\ncompounds (%d):" % len(compounds))
                print(compounds)
                print(f"ar: {ar}")
                print(f"libID: {libID}")
                targetSpecID = pars["targetSpecID"]
                print(f"mixID: {targetSpecID}")

            # this ensures that the total assigned flow < 1.0
            scale = 1 / res.originalWeightY

            assigned = dict(zip(compounds, res.assigned * scale))
            specificCosts = dict(zip(compounds, res.specificCosts))
            absorbed = dict(list(zip(compounds, np.zeros_like(compounds, dtype=float))))
            for i, c in enumerate(compounds):
                if getattr(res, "nPeaksConsidered", None) is None:
                    nPeaks = len(res.compounds[c]["ix"])
                else:
                    nPeaks = res.nPeaksConsidered[i]
                dfCompounds["compound"].append(compoundShortNames.get(c, c))
                dfCompounds["target"].append(targetName)
                dfCompounds["assignment radius"].append(ar)
                dfCompounds["assigned"].append(assigned[c])
                dfCompounds["absorbed"].append(absorbed[c])
                dfCompounds["specific cost"].append(specificCosts[c])
                intScaling = intensityScalings.get(c, None)
                dfCompounds["intensity scaling"].append(intScaling)
                expectedConcentrationFactor = expectedConcentrationFactors.get(
                    targetID, {}
                ).get(c, None)
                dfCompounds["expected concentration factor"].append(
                    expectedConcentrationFactor
                )

                # The concentration factor is calculated as the absolute
                # assigned flow per scan number related to the original
                # compound spectrum intensity per scan number
                cf = (
                    assigned[c]
                    * scanNumberIndividualCompounds
                    / (nScansY * originalWeights[c] * scale)
                )
                af = (
                    absorbed[c]
                    * scanNumberIndividualCompounds
                    / (nScansY * originalWeights[c] * scale)
                )
                if intensityCorrection and c in intensityScalings:
                    cf *= 2 ** (intensityScalings[targetID] - intensityScalings[c])
                    af *= 2 ** (intensityScalings[targetID] - intensityScalings[c])
                else:
                    af = cf = None
                dfCompounds["concentration factor"].append(cf)
                dfCompounds["absorption factor"].append(af)
                if expectedConcentrationFactor:
                    dcf = cf - expectedConcentrationFactor
                    dfCompounds["conc.fctr. mismatch"].append(dcf)
                else:
                    dfCompounds["conc.fctr. mismatch"].append(None)
                dfCompounds["isolated fit"].append(isolated_fit)
                dfCompounds["incremental fit"].append(incrFit)
                dfCompounds["lib"].append(libID)
                dfCompounds["nPeaks"].append(nPeaks)
            totalAbsorbed = res.totalAbsorption * scale
            totalAssigned = res.totalAssignedFlow * scale
            if totalAssigned > 0:
                totalSpecificCost = (
                    sum([specificCosts[c] * assigned[c] for c in compounds])
                    / totalAssigned
                )
            else:
                totalSpecificCost = None
            dfTargets["target"].append(targetName)
            dfTargets["assignment radius"].append(ar)
            dfTargets["assigned"].append(totalAssigned)
            dfTargets["absorbed"].append(totalAbsorbed)
            dfTargets["specific cost"].append(totalSpecificCost)
            dfTargets["isolated fit"].append(isolated_fit)
            dfTargets["lib"].append(libID)
            dfTargets["originalWeightY"].append(res.originalWeightY)

    dfTargets["assigned"] = [float(x) for x in dfTargets["assigned"]]
    dfTargets = pd.DataFrame(dfTargets)
    dfCompounds = pd.DataFrame(dfCompounds)
    dfCompounds["assigned"] = [float(x) for x in dfCompounds["assigned"]]
    dfCompounds["compound"] = [str(x) for x in dfCompounds["compound"]]
    return dfTargets, dfCompounds



import os
import wget

from mcfnmr.config import OLDBFULLSPECTRADIR, OLDB1DTESTDIR, OLDBPEAKSDIR

def prepareOLDBdata():
    for DIR in [OLDB1DTESTDIR, OLDBFULLSPECTRADIR, OLDBPEAKSDIR]:
        if not (DIR).exists():
            os.makedirs(DIR)
            print(f"Created directory '{DIR}'")
            
    baseurl = "https://zenodo.org/records/14888536/files/"
    for s1 in ["I", "II", "III"]:
        for s2 in ["a", "b", "c"]:
            fn = f"nmr_{s1}{s2}_01.txt.gz"
            url = baseurl + fn
            target_fn = OLDBFULLSPECTRADIR / fn
            if not target_fn.exists():
                print(f"Downloading {fn} ...")
                wget.download(url, str(target_fn))
            else:
                print(f"'{target_fn}' already exists. Skipped downloading.")
            print("")
                
    # 1D example
    fn = "nmr_IIa_1H.csv.gz"
    url = baseurl + fn
    target_fn = OLDB1DTESTDIR / fn
    if not target_fn.exists():
        print(f"Downloading {fn} ...")
        wget.download(url, str(target_fn))
    else:
        print(f"'{target_fn}' already exists. Skipped downloading.")
    print("")

    # Compound spectrum peak lists
    fn = "peaklists_compounds.csv"
    url = baseurl + fn
    target_fn = OLDBPEAKSDIR / fn
    if not target_fn.exists():
        print(f"Downloading {fn} ...")
        wget.download(url, str(target_fn))
    else:
        print(f"'{target_fn}' already exists. Skipped downloading.")
    print("")

    print("\nPreparing OLDB data done.\n")


if __name__ == "__main__":
    prepareOLDBdata()



import pandas as pd

from mcfnmr.demodata.loading import loadOLDBCompoundLib, loadOLDB_1D_test 
from mcfnmr.core.mcf import mcf
from mcfnmr.routines.classification import classifyCompounds
from mcfnmr.demodata import mix4
from pprint import pp
from mcfnmr.main import classify_result, save_as_text


def main():
    # load 2D-spectrum inhouse lib as 1D-projection (just for testing!),
    # and include water, DMSO, and TMSP
    lib = loadOLDBCompoundLib(project_1D = True, add_extra = True)
    
    # load 1D spectrum
    # 204 == IIa ("Mixture: D-Glucose, Pimelic acid, 1,2,3-Tricarboxylic acid, Tyrosine (all 30 mM)")
    targetID = "nmr_IIa_1H"
    containedCompounds = mix4 + ["DMSO", "Water", "TMSP"]
    detection_threshold = 1e14
    
    # targetID = "204_1H_solvent_supression"
    # containedCompounds = mix4 + ["DMSO", "Water", "TMSP"]
    # detection_threshold = 1e14
        
    target = loadOLDB_1D_test(targetID, binsize=10)
    
    # # Debug: Plot
    # import matplotlib.pyplot as plt
    # plt.plot(target.coords[:,1], target.weights)
    # import numpy as np
    # plt.plot(target.coords[:,1], np.minimum(0.0, target.weights), color="r")
    # plt.show()
    
    # run recombination
    lib_id = "inhouse_extended"
    res = mcf(target_spectrum=target, library=lib, 
              target_id = target.name, lib_id = lib_id,
              assignment_radius=0.1,
            load=True,
            load_dists=True,)
    
    
    df = classify_result(res, detection_threshold)
    save_as_text(df, "output/mcfresult_test.csv")
    
    nC = len(res.compounds)
    compounds = sorted(res.compounds)
    
    pars = dict(libID=lib_id, targetSpecID=targetID)
    dfCompounds = dict(
        lib=[lib_id]*nC,
        target=[targetID]*nC,
        compound=compounds,
        assigned=res.assigned,
        nPeaks=res.nPeaksConsidered,
        )
    dfCompounds = pd.DataFrame(dfCompounds)
    
    classification = classifyCompounds(dfCompounds, pars, th=detection_threshold, 
                      containedCompounds=containedCompounds, 
                      scale_by_peaknr=False, 
                      verb=False)
    
    del classification["compound_df"]
    print("Result:")
    pp(classification)
    
    
if __name__=="__main__":
    main()





import os
from pathlib import Path

HOMEDIR_WARNINGS = 0

def get_mcfnmr_home():
    global HOMEDIR_WARNINGS
    mcfnmr_home = os.environ.get("MCFNMR_HOME", None)
    if mcfnmr_home is None:
        # home_candidate = Path(__file__).absolute().parent.parent.parent # only for dev
        # set_home_candidate = f"export MCFNMR_HOME='{home_candidate}'"
        home_candidate = os.environ.get("HOME", None) 
        if home_candidate is None:
            # Probably on Windows
            home_candidate = os.environ.get("HOMEPATH", None) 
        if home_candidate is None:
            raise Exception(
                f"Environment variables MCFNMR_HOME and HOME (resp. HOME_PATH) not set."
            )
        home_candidate = Path(home_candidate) / ".mcfnmr"
        if home_candidate.exists():
            if not home_candidate.is_dir():
                raise Exception(
                    f"Environment variable MCFNMR_HOME not set."
                )
            else:
                if HOMEDIR_WARNINGS == 0:
                    print(f"Using MCOM_HOME='{home_candidate}'.")
                    HOMEDIR_WARNINGS += 1
                return home_candidate
        else:
            answer = ""
            answercount, maxcount = 0, 3
            while answer.lower() not in {"y", "yes", "n", "no"} and answercount < maxcount:
                answer = input(f"Environment variable MCFNMR_HOME not set. Use directory '{home_candidate}'? (yes/no)")
            if answer.lower() not in {"y", "yes"}:
                raise Exception(
                    f"Environment variable MCFNMR_HOME not set."
                )
            else:
                os.makedirs(home_candidate)
                print(f"Created directory '{home_candidate}'.")
                print(f"Using MCOM_HOME='{home_candidate}'.")
                return home_candidate
    mcfnmr_home_abs = Path(mcfnmr_home).absolute()
    if not mcfnmr_home_abs.exists():
        raise Exception(
            f"Path '{mcfnmr_home_abs}' (derived from environment variable MCFNMR_HOME='{mcfnmr_home}') doesn't exist."
        )
    subdirs = [str(p.name) for p in mcfnmr_home_abs.iterdir()]
    if not mcfnmr_home_abs.is_dir():
        raise Exception(
            f"Path '{mcfnmr_home_abs}' (derived from environment variable MCFNMR_HOME='{mcfnmr_home}') isn't a directory."
        )
    if (not "tests" in subdirs) or (not "mcfnmr" in subdirs) or (not "data" in subdirs):
        raise Exception(
            f"Path '{mcfnmr_home_abs}' (derived from environment variable MCFNMR_HOME='{mcfnmr_home}') "
            "doesn't contain directories 'tests', 'data', and 'mcfnmr'. Please make sure, that the variable "
            "points to the checkout-directory."
        )
    return mcfnmr_home_abs



import os
import pickle
import numpy as np
from copy import deepcopy
from pprint import pprint

from mcfnmr.config import OUTDIR
from mcfnmr.demodata import case_dict
from mcfnmr.demodata.loading import loadOLDBCompoundLib, loadOLDBMixSpectrum
from mcfnmr.utils.pointspectrum import makeRasterPointSpectrum
from mcfnmr.demodata import OLDBCompoundWeights
from mcfnmr.routines.utils import (
    cache_dir,
    singleRun,
    incrementalMCFResultsUpdate,
    prepareParameters,
    makeResultsDF,
    makeMCFResultFilename,
    updateSavedMCFResult,
)
from mcfnmr.routines.plotting import plotAssignmentStatsOLDB

quantification_cache_dir = cache_dir / "quantification"
if not quantification_cache_dir.exists():
    os.makedirs(quantification_cache_dir)
    print(f"Created directory '{quantification_cache_dir}'")

# Random selection of the 34 compounds to show in barplots
showCompounds = sorted(
    [
        "CinA",
        "Bnz3",
        "HCin",
        "PheB",
        "Bio",
        "Gluc",
        "Man",
        "Met",
        "Pro",
        "Rha",
        "Tyr",
        "Nic",
        "Pim",
        "Van",
    ]
)


def makeStatsFilename(
    targetSpecID,
    binning,
    isolated_fit,
    incrementalFit,
    pointTarget,
    libID,
    assignmentRadiusRange,
    noiseFilterLevel,
    smoothing,
):
    saveStatsDir = os.path.join(OUTDIR, "cache")
    os.makedirs(saveStatsDir, exist_ok=True)
    fn = (
        makeSetupStr(
            targetSpecID,
            binning,
            isolated_fit,
            incrementalFit,
            pointTarget,
            libID,
            assignmentRadiusRange,
            noiseFilterLevel,
            smoothing,
        )
        + ".pickle"
    )
    return os.path.join(saveStatsDir, fn)


def makeSetupStr(
    targetSpecID,
    binning,
    isolated_fit,
    incrementalFit,
    pointTarget,
    libID,
    assignmentRadiusRange,
    noiseFilterLevel,
    smoothing,
):

    isoStr = "" if isolated_fit is None else ("iso" if isolated_fit else "joint")
    incrStr = "incrFit" if incrementalFit else ""
    pointTargetStr = "" if pointTarget is None else ("pt" if pointTarget else "rst")
    noiseFilterStr = (
        "" if noiseFilterLevel is None else "noiseFilter%g" % noiseFilterLevel
    )
    smoothingStr = "_smooth%g" % smoothing
    arRangeStr = "[" + ", ".join(["%g" % ar for ar in assignmentRadiusRange]) + "]"
    return f"{targetSpecID}({binning},{pointTargetStr}){incrStr}{noiseFilterStr}{smoothingStr}_lib{libID}({isoStr})_ar{arRangeStr}"


def saveStats(fn, stats):
    with open(fn, "wb") as f:
        pickle.dump(stats, f)
    print(f"Saved  '{fn}'")


def loadStats(fn):
    with open(fn, "rb") as f:
        stats = pickle.load(f)
    print(f"Loaded '{fn}'")
    return stats


def single_quantification(pars, lib, targetPointSpec, load=True, verb=1):
    savefn = quantification_cache_dir / makeMCFResultFilename(
        pars, pars["targetSpecID"]
    )
    res = singleRun(
        pars=pars, lib=lib, targetPointSpec=targetPointSpec, savefn=savefn, load=load
    )
    return res


def runSingleQuantificationIncremental(pars, arSpan, lib, load=True, return_all=False):
    targetSpec = loadOLDBMixSpectrum(
        specID=pars["targetSpecID"],
        smooth=pars["smoothing"],
        noiseFilterLevel=pars["noiseFilterLevel"],
    )
    targetPointSpec = makeRasterPointSpectrum(
        rasterspec=targetSpec, nbin=pars["binning"]
    )
    targetPointSpec.weights = np.maximum(targetPointSpec.weights, 0.0)

    results = []
    prevResult = None
    prevSavefn = None
    # This reserves assigned sink capacity after each run
    pars["previousARSteps"] = []
    pars["reservedSinkCapacity"] = np.zeros_like(targetPointSpec.weights)
    for ar in arSpan:
        pars = deepcopy(pars)
        pars["assignmentRadius"] = ar
        savefn = quantification_cache_dir / makeMCFResultFilename(
            pars, pars["targetSpecID"]
        )
        res = singleRun(
            pars=pars,
            lib=lib,
            targetPointSpec=deepcopy(targetPointSpec),
            savefn=savefn,
            load=load,
        )
        loaded_res = res.inflowsY is None
        # Update sequence memory
        pars["previousARSteps"].append(ar)
        if (
            not loaded_res
            and np.abs(np.round(sum(res.inflowsY) - sum(res.assigned), 6)) > 0.0
        ):
            raise Exception(
                "runSingleQuantificationIncremental() total target inflows must equal sum of individual assignments!"
            )
        if not loaded_res:
            # Add up assigned flow from previous runs
            incrementalMCFResultsUpdate(res, res.inflowsY, prevResult)
        pars["reservedSinkCapacity"] = res.reservedSinkCapacity
        # Delete inflowsY and residualsY from result to save storage.
        # We will only need the reserved capacity at the next
        # larger ar-value to extend an incremental series
        res.inflowsY = None
        res.residualsY = None
        updateSavedMCFResult(res, savefn=savefn)
        if prevResult:
            # reservedCapacity not needed for non-maximal ar
            prevResult.reservedSinkCapacity = None
            updateSavedMCFResult(prevResult, savefn=prevSavefn)
        results.append({"res": res, "pars": pars})
        prevSavefn = savefn
        prevResult = res
    if return_all:
        return results
    else:
        return results[-1]


def runSingleQuantification(pars, ar, lib, load=True):
    targetSpec = loadOLDBMixSpectrum(
        specID=pars["targetSpecID"],
        smooth=pars["smoothing"],
        noiseFilterLevel=pars["noiseFilterLevel"],
    )
    targetPointSpec = makeRasterPointSpectrum(
        rasterspec=targetSpec, nbin=pars["binning"]
    )
    pars = deepcopy(pars)
    pars["assignmentRadius"] = ar
    savefn = quantification_cache_dir / makeMCFResultFilename(
        pars, pars["targetSpecID"]
    )
    res = singleRun(
        pars=pars,
        lib=lib,
        targetPointSpec=deepcopy(targetPointSpec),
        savefn=savefn,
        load=load,
    )
    result = {"res": res, "pars": pars}
    return result


def main(cfg):
    assert cfg["task"] == "quantification"
    assert cfg["lib"].lower() == "inhouse"
    mix = cfg["mix"].split(".")
    assert len(mix) == 2
    series, experiment = mix
    assert series in ["I", "II", "III"]
    assert experiment in ["a", "b", "c"]
    assert cfg["setup"] in ["A", "B", "C"]

    # For incremental assignment
    arSpan = np.array(
        [
            0.025,
            0.05,
            0.075,
            0.1,
            0.125,
            0.15,
            0.175,
            0.2,
            0.225,
            0.25,
            0.3,
            0.35,
            0.4,
            0.5,
        ]
    )

    # OLDB compound lib
    lib = loadOLDBCompoundLib()
    setupSpec = case_dict[cfg["setup"]][cfg["mix"]]
    isolated_fit = setupSpec["isolated_fit"]
    incrementalFit = setupSpec["incrementalFit"]
    pointTarget = setupSpec["pointTarget"]

    # All individual compounds with concentrations 30mM
    libID = "OLDBcompoundLib"
    targetSpecID = series + experiment + "_01"
    binning = (2, 4)
    noiseFilterLevel = None
    smoothing = 0
    pars = prepareParameters(
        isolated_fit=isolated_fit,
        targetSpecID=targetSpecID,
        pointTarget=pointTarget,
        libID=libID,
        binning=binning,
        specDomain="inhouse",
        incrementalFit=incrementalFit,
        noiseFilterLevel=None,
        smoothing=0,
        workingDir=cfg["outdir"],
        recompute=cfg.get("recompute", True),
    )

    load = not pars["recompute"]
    # Run
    if incrementalFit:
        ars = np.append(arSpan[arSpan < cfg["ar"]], cfg["ar"])
        res = runSingleQuantificationIncremental(pars, arSpan=ars, lib=lib, load=load)
    else:
        res = runSingleQuantification(pars, ar=cfg["ar"], lib=lib, load=load)
        ars = [cfg["ar"]]

    # This assumes that all weight of X is in the target regions
    dfTargets, dfCompounds = makeResultsDF(
        [res], intensityCorrection=True, originalWeights=OLDBCompoundWeights
    )

    setupStr = makeSetupStr(
        targetSpecID,
        binning,
        isolated_fit,
        incrementalFit,
        pointTarget,
        libID,
        ars,
        noiseFilterLevel,
        smoothing,
    )
    figs_dir = OUTDIR / "figs"
    stats = plotAssignmentStatsOLDB(
        dfCompounds,
        dfTargets,
        libID,
        targetSpecID,
        showCompounds,
        outdir=figs_dir,
        show=False,
        setupStr=setupStr,
    )
    print(f'error_stats["{targetSpecID}"] = \\')
    pprint(stats["errors"])
    fn = makeStatsFilename(
        targetSpecID,
        binning,
        isolated_fit,
        incrementalFit,
        pointTarget,
        libID,
        ars,
        noiseFilterLevel,
        smoothing,
    )
    saveStats(fn, stats)


if __name__ == "__main__":
    # Test run
    cfg = dict(
        lib="InHouse",
        mix="I.a",
        recompute=False,
        task="quantification",
        setup="C",
        ar=0.15,
        outdir=cache_dir,
    )
    main(cfg)



import os
import gzip
import pickle
import pandas as pd
import numpy as np
import xml.etree.ElementTree as ET
from pprint import pp
from warnings import warn

from mcfnmr.config import (
    PROCESSED_UCSFDIR,
    OLDBPEAKSDIR,
    SMARTMINER_DATADIR,
    MM_PEAKLISTDIR,
    MM_LIBDIR,
    MM_MIXLISTSDIR,
    OLDBFULLSPECTRADIR,
    DEBUG, OLDB1DTESTDIR,
)
from mcfnmr.utils.rasterspectrum import RasterSpectrum
from mcfnmr.utils.pointspectrum import PointSpectrum
from mcfnmr.demodata import metaboMinerLibNames, metaboMinerLibNameDict, correctSubstrateNames
from mcfnmr.utils.loading import parseBrukerSpectrum
from copy import deepcopy

aliases = {
    # From MM mixtures
    "Oxoglutarate": "2-Oxoglutarate",
}

userprovided_HMDB_IDs = {
    # From MM mixtures
    "Urea": "294",
    "AMP": "45",  # include
    # From Drosophila annotation
    "Beta-alanine": "56",
    "Glucose-6-P": "1401",
    # From Drosophila annotation
    "DSS": None,  # bmse000795
    # From COLMAR recognitions in N926
    "L-Iditol": 11632,
    "L-Threitol": 2994,
    "Sulfoacetic acid": 258590,
    "1,2_Propanediol": 1881,
    "D-Pinitol": 34219,
    "4-Methyl-2-oxovaleric acid": None,  # bmse000383
    "N-Acetyl-glucosamine 1-phosphate": 1367,
    "D-Glucosaminic acid": 341308,
    "Fructose 1-phosphate": 1076,
    "compound_unknown": None,
    "Isovaleraldehyde": 6478,
    "Neopterin": 845,
    "2-Oxohexane": 5842,
    "Cadaverine": 2322,
    "alpha-Ketoglutaric acid": 208,
    "Adonitol": 508,
    "6-Aminohexanoic acid": 1901,
    "Hydroxyphenylacetylglycine": 735,
    "Gulonic acid": 3290,
    "4-Hydroxyphenylglycine": 244973,
    "Galactonolactone": 2541,
    "Stachyose": 3553,
    "cis-4-Octenedioic acid": 4982,
    "L-Gulonolactone": 3466,
    "D-galactono-1,4-Lactone": 2541,
    "delta-Hexanolactone": 453,
    "D-Xylonate": 59750,
    "scyllo-Inositol": 6088,
    "alpha-D-Glucose-1-phosphate": 1586,
    "meso-Erythritol": 2994,
    "alpha,epsilon-Diaminopimelic acid": 1370,
    "Arbutin": 29943,
    "Argininosuccinic acid": 52,
    "gamma-Aminobutyric acid": 112,
    "Trimethyl phosphate": 259237,
    "N-Acetyl-D-glucosamine-6-phosphate": 1062,
    "Chitosan": 3404,
    "Dehydroascorbic acid": 1264,
    "O-phosphothreonine": 11185,
    "Maltotetraose": 1296,
    "Xylulose": 751,
    "Tetramethylammonium": None,  # bmse000780
    "muco-Inositol": 62138,
    "DL-alpha-Glycerol phosphate": 126,
    "Hexanoic acid": 535,
    "N-Carbamoyl-L-glutamic acid": 15673,
    "Malic acid": 156,
    # From SMART-Miner recognitions in N926
    "beta-Gentiobiose": 341292,
    "Unknown": None,
    "Propylene glycol": 1881,
    "Sodium nalidixic acid": 14917,
    "Nalidixic acid sodium salt": 14917,
    "Muramic acid": 3254,
    "Citicoline": 1413,
    "Allothreonine": 4041,
    "7-Methyladenine": 11614,
    "4-Hydroxyproline": 725,
    "3-Pyridylacetic acid": 1538,
    "3-Methylcrotonaldehyde": 12157,
    "3-Methyl-2-butenal": 12157,
    "3-Methyladenine": 11600,
    "3-Mercaptopyruvic acid": 1368,
    "2-Heptanone": 3671,
    "L-2-Aminobutyric acid": 452,
    "2-Amino-5-ethyl-1,3,4-thiadiazole": None,  # bmse000176
    "1-Methyladenine": 11599,
    "p-Cresol": 1858,
    "Mesaconic acid": 749,
    "L-Homocysteic acid": 2205,
    "Citramalic acid": 426,
    "Arecoline hydrobromide": 30353,
    "4-Heptanone": 4814,
    "4-Aminoantipyrine": 246350,
}


def buildHMDBID2NameMap(verb=1):
    name2id = buildName2HMDBIDMap(verb=verb)
    id2name = dict()
    for name in sorted(name2id.keys()):
        id2name[name2id[name]] = name
    return id2name


def buildName2HMDBIDMap(verb=DEBUG):
    metaboDBFile = MM_LIBDIR / "hsqc.xml"
    et = ET.parse(metaboDBFile)
    libElems = dict([(elem.attrib["Name"], elem) for elem in list(et.getroot())])

    # Use full lib for compound name lookup
    lib = libElems[metaboMinerLibNames[0]]

    # construct mapping name → HMDB ID
    print("Building HMDB-ID list ...")
    map_name2HMDB = {name: str(ID) for name, ID in userprovided_HMDB_IDs.items()}
    for compound in list(lib):
        name = compound.attrib["Name"]
        HMDB_ID = compound.attrib["HMDB_ID"]
        if len(HMDB_ID) == 0:
            warn("No HMDB ID for '%s'!" % name)
            continue
        else:
            HMDB_ID = str(int(HMDB_ID[4:]))
        if name in map_name2HMDB:
            smaller = min(HMDB_ID, map_name2HMDB[name])
            larger = max(HMDB_ID, map_name2HMDB[name])
            if larger != smaller:
                warn(
                    f"\nDuplicate entry for '{name}'! Using smaller ID ({smaller}) and discarding '{larger}'"
                )
            HMDB_ID = smaller
        map_name2HMDB[name] = HMDB_ID
        if name in aliases:
            alias = aliases[name]
        elif name[-3:] == "ate":
            alias = name[:-3] + "ic acid"
        elif name[-7:] == "ic acid":
            alias = name[:-7] + "ate"
        else:
            alias = None
        if verb > 0:
            print(f"   {name} → {HMDB_ID}")

        # Ignore compounds with userprovided_HMDB_IDs[alias]==None
        if alias in userprovided_HMDB_IDs and userprovided_HMDB_IDs[alias] is None:
            print(f"Ignoring alias '{alias}' for {name} with ID '{HMDB_ID}'")
            alias = None

        if alias:
            map_name2HMDB[alias] = HMDB_ID
            if verb > 0:
                print(f"   {alias} → {HMDB_ID}")
    return map_name2HMDB


def loadSMARTMINERBenchmarks():
    # Load Benchmarks from SMART-Miner Publication
    fn = os.path.join(SMARTMINER_DATADIR, "benchmarktable.csv")
    tab = pd.read_csv(fn, index_col=False)
    tab.columns = [col.strip() for col in tab.columns]
    print(tab)

    df = {
        "mixID": [],
        "libID": [],
        "matchType": [],
        "assignmentRadius": [],
        "binning": [],
        "absorptionCost": [],
        "precision": [],
        "recall": [],
        "F1 score": [],
        "cutoffFlag": [],
        "hueID": [],
        "method": [],
        "FP": [],
        "TP": [],
        "FN": [],
    }

    parseMethods = ["Metabominer", "COLMAR-HSQC", "SMART-Miner"]
    n = len(tab["Method"])
    for i in range(n):
        method = tab["Method"][i]
        if method not in parseMethods:
            print(f"Skipping method '{method}'")
            continue
        elif method == "Metabominer":
            # Unify spelling
            method = "MetaboMiner"
        df["mixID"].append(tab["Sample"][i].strip())
        df["libID"].append("Biofluid ( all )")
        df["recall"].append(tab["Recall"][i])
        df["precision"].append(tab["Precision"][i])
        df["F1 score"].append(tab["F-score"][i])
        df["FP"].append(tab["FP"][i])
        df["FN"].append(tab["FN"][i])
        df["TP"].append(tab["TP"][i])
        df["hueID"].append(None)
        df["method"].append(method)

    for k in list(df.keys()):
        if len(df[k]) == 0:
            df[k] = None
    df = pd.DataFrame(df)
    return df


def loadMetaboMinerSpectrum(mixID, denoising_pars={}):
    fn = PROCESSED_UCSFDIR / (mixID + ".pickle")
    spec = parseMetaboMinerDataFile(fn, denoising_pars=denoising_pars)
    return spec


def parseMetaboMinerDataFile(fn, denoising_pars={}):
    with open(fn, "rb") as f:
        data = pickle.load(f)
    print(f"Loaded data from '{fn}'")
    fullData = data["matrix"]
    header = data["header"]

    if DEBUG > 1:
        print("Header:")
        pp(header)

    spec = RasterSpectrum(
        fullData, header, denoising_pars=denoising_pars, name=fn.name[:-7]
    )
    return spec


def loadMMMixPointSpectra():
    mixPeaks = {}
    for fn in (MM_PEAKLISTDIR).iterdir():
        fn_split = fn.name.split(".")
        specID = ".".join(fn_split[:-1])
        ext = fn_split[-1]
        mixID, specType = specID.split("_")
        if (specType != "hsqc") or ext != "txt":
            continue
        data = np.genfromtxt(fn, dtype=float, skip_header=1)
        coords = data[:, :2]
        weights = data[:, 2]
        spec = PointSpectrum(coords, weights)
        mixPeaks[mixID] = spec
        print(f"Loaded point spectrum for mix '{mixID}' with %d peaks." % len(weights))
    return mixPeaks


def loadMMMixtureLists():
    files = [f for f in (MM_MIXLISTSDIR).iterdir() if f.name[-9:] == "_HMDB.txt"]
    mixLists = {}
    for fn in files:
        mixID = fn.name[:-9]
        with open(fn, "r") as f:
            mixLists[mixID] = [l.strip() for l in f.readlines() if l.strip()]
        print(f"Loaded mix '{mixID}' content list of length %d" % len(mixLists[mixID]))
    return mixLists


def loadMMLibrary(libName="metabominer", verb=False):
    if libName.lower() == "metabominer" or libName is None:
        # Default MM lib is full "Biofluid (all)" library
        libName = metaboMinerLibNames[0]
    elif libName.lower().find("metabominer") > -1:
        libName = metaboMinerLibNameDict["-".join(libName.split("-")[1:]).lower()]
        
    metaboDBFile = MM_LIBDIR / "hsqc.xml"
    et = ET.parse(metaboDBFile)
    libElems = dict([(elem.attrib["Name"], elem) for elem in list(et.getroot())])
    lib = libElems[libName]

    print("Loading MM lib '%s' containing %d compounds..." % (libName, len(lib)))

    peakLists = {}
    for compound in list(lib):
        name = compound.attrib["Name"]

        # Ignore userspecified IDs
        if name in userprovided_HMDB_IDs and userprovided_HMDB_IDs[name] is None:
            assert(False)
            continue

        HMDB_ID = compound.attrib["HMDB_ID"]
        if len(HMDB_ID) == 0:
            warn("No HMDB ID for '%s'!" % name)
        else:
            HMDB_ID = str(int(HMDB_ID[4:]))
        if HMDB_ID in peakLists:
            raise Exception(
                f"\nDuplicate peaklist entry for '{HMDB_ID}' in lib '{libName}'!"
            )
        pls = list(compound)
        if len(pls) > 1:
            warn("   Found %d peaklists for '%s'!? (using first...)" % (len(pls), name))
            raise Exception(f"Several peaklists for '{HMDB_ID}' in lib '{libName}'!")
        elif len(pls) == 0:
            warn("   Found no peaklists for '%s'!?" % (name))
            raise Exception(f"Missing peaklist for '{HMDB_ID}' in lib '{libName}'!")
            # continue
        pl = pls[0]
        if verb:
            print(f"{name} (HMDB-ID {HMDB_ID}):")
            print("   Number of peaks: %d" % len(pl))

        # Coords and weights (latter are not provided by MetaboMiner)
        coords = []
        for peak in pl:
            x, y = peak.attrib["X_PPM"], peak.attrib["Y_PPM"]
            coords.append((float(y), float(x)))
        peakLists[HMDB_ID] = PointSpectrum(
            weights=np.ones(len(coords)), coords=coords, name=name
        )
        if verb:
            print("   peaks:", peakLists[HMDB_ID]["coords"])
    # Manual removal of duplicate entries (which have the same spectrum)
    if ("190" in peakLists) and ("1311" in peakLists):
        # Lactic Acid vs D-Lactic Acid
        print("Removing duplicate entry for D-Lactic Acid (HMDB 1311)")
        del peakLists["1311"]

    return peakLists


def loadOLDBMixSpectrum(specID, smooth, noiseFilterLevel):
    specFiles = sorted(
        [fn for fn in (OLDBFULLSPECTRADIR).iterdir() if fn.name[-7:] == ".txt.gz"]
    )
    specFiles = {"_".join(fn.name.split(".")[0].split("_")[1:]): fn for fn in specFiles}

    pp(specFiles)
    if specID not in specFiles:
        raise Exception(
            f"Couldn't find mixture spectrum '{specID}'. Might need to download it to {OLDBFULLSPECTRADIR}."
        )
    fn = specFiles[specID]
    fullData, header = parseOLDBTextFile(fn)
    spec = RasterSpectrum(
        fullData,
        header,
        name=specID,
        denoising_pars=dict(smoothRadius=smooth, noiseFilterLevel=noiseFilterLevel),
    )
    return spec


def parseOLDBTextFile(fn):
    return parseBrukerSpectrum(fn)


def projectLibTo1D(lib, axis=1):
    compounds = sorted(lib)
    lib1D = dict()
    for c in compounds:
        spec2D = lib[c]
        spec1D = deepcopy(spec2D)
        nc = len(spec2D.coords)
        dim = len(spec1D.coords[0,:])
        for i in range(dim):
            if i == axis:
                continue
            else:
                spec1D.coords[:,i] = np.zeros(nc)
        lib1D[c] = spec1D
    return lib1D


def add_1D_extra_cpds_oldb(lib):
    lib["Water"] = PointSpectrum(coords = np.array([[0.0, 3.33]]), 
                                 weights = np.array([1.0]), 
                                 name = "water")
    lib["DMSO"] = PointSpectrum(coords = np.array([[0.0, 2.54]]), 
                                 weights = np.array([1.0]), 
                                 name = "DMSO")
    lib["TMSP"] = PointSpectrum(coords = np.array([[0.0, 0.0]]), 
                                 weights = np.array([1.0]), 
                                 name = "TMSP")
    

def loadOLDBCompoundLib(project_1D = False, add_extra = False):
    # "add_extra" adds spectra for Water, DMSO, and TMSPD
    # "project_1D" projects spectra to H-axis
    if add_extra and not project_1D:
        warn("loadOLDBCompoundLib() ignoring add_extra (only implemented for 1D spectra)")
        add_extra = False
    peakData = loadOLDBPeakData()
    df = peakData["peaklists_compounds"]
    substances = sorted(set(df["Substance"]))
    lib = {}
    for s in substances:
        # print("s = '%s'"%s)
        dfs = df[df["Substance"] == s]
        weights = []
        coords = []
        for i, f1, f2 in zip(dfs["Integral [abs]"], dfs["ν(F1) [ppm]"], dfs["ν(F2) [ppm]"]):
            # print("   Peak at (%g, %g) with weight: %g"%(f1, f2, i))
            weights.append(i)
            coords.append((f1, f2))
        s = correctSubstrateNames.get(s,s)
        spec = PointSpectrum(coords=coords, weights=weights, name=s)
        lib[s] = spec
    if project_1D:
        # Assuming H-coordinate is second (i.e., axis 1)
        lib = projectLibTo1D(lib, axis=1)
        if add_extra:
            add_1D_extra_cpds_oldb(lib)
    return lib



def bin_1D(weights, coords, binsize):
    n = len(weights)
    dim = coords.shape[1]
    n_binned = int(np.floor(np.round(n/binsize,10)))
    weights_binned = np.zeros(n_binned)
    coords_binned =  np.zeros((n_binned, dim))
    
    for i in range(n_binned):
        i0, i1 = i*binsize, (i+1)*binsize
        weights_binned[i] = np.sum(weights[i0:i1])
        coords_binned[i, 1] = (coords[i0, 1] + coords[i1, 1])/2
    return weights_binned, coords_binned


def loadOLDB_1D_test(target_spec_name, binsize):
    fn = OLDB1DTESTDIR / f"{target_spec_name}.csv.gz"
    with gzip.open(fn, "rb") as f:
        df = pd.read_csv(f)
    H_shifts = np.array(df["chem. Shift/ppm"])
    weights = np.maximum(0.0, np.array(df["rel. intensity"]))
    coords = np.array([np.zeros_like(H_shifts), H_shifts]).T
    weights, coords = bin_1D(weights, coords, binsize)
    spec = PointSpectrum(coords, weights, name=f"1D_test_binsize{binsize}")
    return spec
    

def loadOLDBPeakData():
    # Load all peak data from OL experiments
    data_files = [fn for fn in os.listdir(OLDBPEAKSDIR) if fn[-3:] == "csv"]
    data = {}
    for fn in data_files:
        dataID = fn[:-4]
        print("Loading data '%s' ..." % dataID)
        fullfn = OLDBPEAKSDIR / fn
        data[dataID] = pd.read_csv(fullfn, sep="\t", header=0)
    return data



#! /usr/bin/python

import os
import sys
import tomllib
import argparse
import time
from pathlib import Path
from pprint import pprint
from argparse import Namespace

import mcfnmr
from mcfnmr.utils import spec2csv
from mcfnmr.utils.system import get_mcfnmr_home
from mcfnmr.routines import (
    scan,
    comparison_classification,
    comparison_quantification,
    classification,
    quantification,
    analysis_plasma_samples, test1D,
)

sys.path.append(os.environ["MCFNMR_HOME"])
os.environ["PYTHONPATH"] = (
    os.environ.get("PYTHONPATH", "") + ":" + os.environ["MCFNMR_HOME"]
)


def parse_args():
    arg_parser = argparse.ArgumentParser(
        prog="mcfNMR test runner",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    arg_parser.add_argument(
        "--outdir",
        "-o",
        default=None,
        dest="outdir",
        help="Output base directory for generated files.",
        required=True,
        type=str,
    )
    arg_parser.add_argument(
        "--config",
        "-c",
        default=None,
        dest="config",
        help="Configuration and parameter file for the run",
        required=True,
        type=str,
    )
    return arg_parser.parse_args()


def run(args):
    MCFNMR_HOME = get_mcfnmr_home()

    # Run directory
    rundir = Path(os.getcwd())
    print(f"\nWorking directory: {rundir}")

    # Load config
    fn = Path(args.config)
    with open(fn, "rb") as file:
        cfg = tomllib.load(file)
    print("\nconfig:")
    pprint(cfg)

    # Output directory
    outdir = Path(args.outdir)
    if outdir != outdir.absolute():
        outdir = rundir / outdir
    if not outdir.exists():
        os.makedirs(outdir)
    print(f"\nWriting output to: {outdir}")
    cfg["outdir"] = outdir

    # Data directory
    datadir = MCFNMR_HOME / "data"
    print(f"\nData directory: {datadir}")
    if not datadir.exists():
        raise Exception(f"Couldn't find data directory '{datadir}'")
    cfg["datadir"] = datadir

    # Debug
    # Whether to recompute all parts of the result
    # or use cache for the run (speeds up a lot, but
    # should be turned off for full testing)
    # cfg["recompute"] = True
    # cfg["recompute"] = False

    # Run task
    start_time = time.time()
    if cfg["task"] == "classification":
        classification.main(cfg)
    elif cfg["task"] == "quantification":
        quantification.main(cfg)
    elif cfg["task"] == "scan":
        scan.main(cfg)
    elif cfg["task"] == "test1D":
        test1D.main()
    elif cfg["task"] == "comparison_classification":
        comparison_classification.main(cfg)
    elif cfg["task"] == "comparison_quantification":
        comparison_quantification.main(cfg)
    elif cfg["task"] == "plasma_sample_analysis":
        analysis_plasma_samples.main(
            load=not cfg.get("recompute", False), 
            plot=cfg.get("plot", False),
            libName = cfg.get("lib", "MetaboMiner-all"),
        )
    elif cfg["task"] == "spec2csv":
        if cfg.get("help", False):
            sys.argv = ["spec2csv", "--help"]
            spec2csv.main()
        else:
            args = Namespace(
                infile=str(MCFNMR_HOME / cfg["target"]),
                outfile=str(outdir / cfg["outfile"]),
                bindims=cfg["bindims"],
                xrange=None,
                yrange=None,
                irange=None,
                override=False,
                sqrt=False,
                peaklist=cfg.get("peaklist", False),
            )
            args, msg = spec2csv.process_args(args)
            if args is None:
                raise Exception(f"Problem with test config: '{msg}'")
            spec2csv.run(args)
    elif cfg["task"] == "mcfNMR":
        if cfg.get("help", False):
            sys.argv = ["mcfNMR", "--help"]
            mcfnmr.main()
        else:
            args = Namespace(
                config=get_mcfnmr_home() / cfg["config"],
                outdir=outdir,
            )
            config = mcfnmr.load_config(args)
            for k, v in cfg.items():
                if k in ["outdir", "config"]:
                    continue
                config[k] = v
            mcfnmr.run(config)
    else:
        raise Exception(f"Unknown task '{cfg['task']}'")
    elapsed = time.time() - start_time
    print(f"\n Total run time: {elapsed} secs")


def main(config_fn=None):
    if config_fn is None:
        print("Test run ...")
        args = parse_args()
        print("\nargs:")
        print(args)
    else:
        args = Namespace(
            config=config_fn,
            outdir=get_mcfnmr_home() / "output" / "tmp",
        )
    run(args)


if __name__ == "__main__":
    # Debug example
    # main(get_mcfnmr_home() / "tests" / "user_interface" / "mcfNMR" / "mcfNMR_N988_grid_incremental" / "test_config.toml")
    main()
